package edu.chl.hajo.monopoly.event;

import java.beans.PropertyChangeEvent;
import java.beans.PropertyChangeListener;
import java.util.ArrayList;
import java.util.HashSet;
import java.util.List;
import java.util.Map;
import java.util.Set;
import java.util.concurrent.ConcurrentHashMap;
import java.util.logging.Logger;

import edu.chl.hajo.monopoly.constants.GameEvents;
import edu.chl.hajo.monopoly.constants.StateChanges;

/**
 * All classes firing events and listening for events uses this.
 * 
 * We have tried to refactor out the Observer Pattern from the
 * model.
 * Now all connections can be done outside the model (se GUIBuider).
 * 
 * Usage connecting;
 * Static connections known at program start could be
 * set outside any involved object (especially GUI objects)
 * using connect(dest, source)
 * 
 * Dynamic connections must connect them selves (pop up windows etc.)
 * possible using connect(this) if no source (accepting broadcast calls).
 * If source is available use possible connect(dest, source).
 * 
 * Usage send and broadcast;
 * send will send from one source to all registered destinations
 * (i.e. the first parameter)
 * broadcast will send to all registered destinations
 * 
 * @author hajo
 *
 */
public final class EventBus {
	private Logger log = Logger.getLogger(this.toString());
	// Used as source when no source given, in connect(dest) method;
	private final Object def = new Object();
	private boolean on = true;

	private static EventBus instance;

	public static synchronized EventBus getInstance() {
		if (instance == null) {
			instance = new EventBus();
		}
		return instance;
	}
	
	/*
	 * Possible to turn off (testing)
	 */
	public void on( boolean b){
		this.on = b;
	}

	private EventBus() {
		connections.put(def, new ArrayList<PropertyChangeListener>());
	}

	/*
	 * ConcurrentHashMap is crucial else we'll get a ConcurrentModificationException
	 * if connect() is called from broadcast()
	 */
	private Map<Object, List<PropertyChangeListener>> connections = 
		new ConcurrentHashMap<Object, List<PropertyChangeListener>>();

	/**
	 * Connect a source to possible many destinations
	 */
	public void connect(PropertyChangeListener dest, Object source) {
		if (dest == null || source == null) {
			throw new IllegalArgumentException("Sourse or destination null");
		}
		List<PropertyChangeListener> ls = connections.get(source);
		if (ls == null) {
			ls = new ArrayList<PropertyChangeListener>();
			connections.put(source, ls);
		}
		if (!ls.contains(dest)) {
			ls.add(dest);
		}
		// log.info("Connected dest " + dest.getClass().getSimpleName() +
		// " to source " + source.getClass().getSimpleName());
	}

	/**
	 * Connect for receiving broadcasts
	 */
	public void connect(PropertyChangeListener dest) {
		if (dest == null) {
			throw new IllegalArgumentException("Destination null");
		}
		List<PropertyChangeListener> ls = connections.get(def);
		if (!ls.contains(dest)) {
			ls.add(dest);
		}
		// log.info("Connected dest " + dest.getClass().getSimpleName() +
		// " to source " + source.getClass().getSimpleName());
	}

	// TODO public void disconnect(PropertyChangeListener dest)

	/**
	 * Send from one source to all connected destinations
	 * for that source 
	 */
	public void send(Object source, String name, Object oldValue,
			Object newValue) {
		if( !on ){
			return;
		}
		List<PropertyChangeListener> ps = connections.get(source);
		if (ps == null || ps.size() == 0) {
			throw new IllegalStateException("No connection for " + source);
		}
		PropertyChangeEvent e = new PropertyChangeEvent(source, name, oldValue,
				newValue);
		for (PropertyChangeListener p : ps) {
			p.propertyChange(e);
			log.info(name.toString() + " from "
					+ source.getClass().getSimpleName() + " to "
					+ p.getClass().getSimpleName());
		}
	}

	/**
	 * Sent to every connected destination 
	 */
	public void broadcast(String name, Object oldValue, Object newValue) {
		if( !on ){
			return;
		}
		log.info(name.toString());
		Set<PropertyChangeListener> uniqueListeners = new HashSet<PropertyChangeListener>();
		for (List<PropertyChangeListener> ps : connections.values()) {
			for (PropertyChangeListener p : ps) {
				// Remove duplicates listeners 
				uniqueListeners.add(p);
			}
		}
		for(PropertyChangeListener p : uniqueListeners){
			// Can't have null as source, put this in (not used)
			PropertyChangeEvent e = new PropertyChangeEvent(this, name, oldValue,
					newValue);
			p.propertyChange(e);
		}
	}

	public void send(Object source, StateChanges name, Object oldValue,
			Object newValue) {
		send(source, name.toString(), oldValue, newValue);
	}

	public void send(Object source, GameEvents name, Object oldValue,
			Object newValue) {
		send(source, name.toString(), oldValue, newValue);
	}

	public void broadcast(GameEvents name, Object oldValue, Object newValue) {
		broadcast(name.toString(), oldValue, newValue);
	}

	public void broadcast(StateChanges name, Object oldValue, Object newValue) {
		broadcast(name.toString(), oldValue, newValue);
	}
}
